@@ -95,30 +95,30 @@ class DSI():
             'hidedsp': False, # If True, hides the surface/boundary
             
             # ----- Plot Format ----- #
-            'czorder': False,         # computed_zorder for 3D plots settings
-            'fs': (6, 4),            # Figure size
-            'bw': False,             # If True, use black-white template
-            'alpha': 0.45,           # Transparency of points
-            'font_size': 10,         # Font size
+            'czorder': False,     # computed_zorder for 3D plots settings
+            'fs': (6, 4),         # Figure size
+            'bw': False,          # If True, use black-white template
+            'alpha': 1,           # Transparency of points
+            'font_size': 14,      # Font size
             # Color bar and map
             'cbarloc': 'right',   # Colobar location
             'cbaror': 'vertical', # Colorbar orientation
-            'mycmap': None,       # Use your own cmap (input name of cmap as str)
+            'mycmap': 'viridis',  # Use your own cmap (input name of cmap as str)
             # Satisfied samples
-            'satlabel': 'Satisfied', # Satisfied samples label
-            'satcolor': 'g',         # Satisfied samples color
-            'satmarker': 'o',        # Marker of satisfied points
-            'satfill': 'g',          # Marker fill color of satisfied points
-            'satzorder': 5,          # Decides which level to be plotted on
+            'satlabel': 'Sat',     # Satisfied samples label
+            'satcolor': '#FF9000', # Satisfied samples color
+            'satmarker': '.',      # Marker of satisfied points
+            'satfill': '#FF9000',  # Marker fill color of satisfied points
+            'satzorder': 5,        # Decides which level to be plotted on
             # Violated samples
-            'violabel': 'Violated',  # Violated samples label
-            'viocolor': 'r',         # Violated samples color
-            'viomarker': 'o',        # Marker of violated points
-            'viofill': 'r',          # Marker fill color of violated points
-            'viozorder': 5,          # Decides which level to be plotted on
+            'violabel': 'Vio',     # Violated samples label
+            'viocolor': '#005DC1', # Violated samples color
+            'viomarker': '.',      # Marker of violated points
+            'viofill': '#005DC1',  # Marker fill color of violated points
+            'viozorder': 5,        # Decides which level to be plotted on
             # Legend
-            'legloc': 'best',        # Legend location
-            'framealpha': 0.8,       # Legend box transparency
+            'legloc': 'upper right', # Legend location
+            'framealpha': 1,         # Legend box transparency
             'legendzorder': 100,     # Decides which level to be plotted on
             # 3D view
             'elev': 20,              # Elevation of 3D plot
@@ -131,7 +131,7 @@ class DSI():
             # ----- Design Space Parameters ----- #
             'dsplabel': 'DSp',   # Label of surface/boundary
             'dspcolor': 'black', # Color of the surface/boundary (both)
-            'dspwidth': 3,       # Thickness of the boundary (2D)
+            'dspwidth': 4,       # Thickness of the boundary (2D)
             'dspstyle': '-',     # Line style of the boundary (2D)
             'dspalpha': 0.2,     # Transparency of surface/boundary (3D)
             'dspzorder': 20,     # To make sure it is plotted ontop of the samples
@@ -140,23 +140,35 @@ class DSI():
             'step_change': 1,    # Step change of expanding AOR in percent
             'noplabel': 'NOP',   # Normal Operating Point label for legend
             'nopmarker': 'x',    # Nominal operating point marker style
-            'nopwidth':   3,     # Nominal operating point marker thickness
+            'nopwidth':   4,     # Nominal operating point marker thickness
             'nopcolor': 'black', # Nominal operating point marker color
-            'nopsize':  100,     # Nominal operating point marker size
+            'nopsize':  150,     # Nominal operating point marker size
             'nopzorder': 10,     # To make sure it is plotted ontop of the samples
 
             # ----- AOR Parameters ----- #
             'aorlabel': 'AOR',   # Uniform Proven Acceptable Range
             'aorstyle': '--',    # AOR boundary line style
             'aorcolor': 'black', # AOR boundary line color
-            'aorwidth': 3,       # AOR boundary line width
+            'aorwidth': 4,       # AOR boundary line width
             'aorzorder': 10,     # To make sure it is plotted ontop of the samples
+            'AORlb': 0,          # Bisection params (lower bound)
+            'AORub': 1,          # Bisection params (upper bound)
+            'AORtol': 0.001,     # Bisection params - tolerance
+            'AORmaxiter': 50,    # Bisection params - max iterations
+            'AORprintF': False,  # Print bisection output
             
             # ----- Hull Parameters ----- #
-            'a': None, # Alpha value -> at large alpha,
-                       # hull becomes convex. if set to 'critical', MATLAB finds 
-                       # the smallest one. if None: use mean of bounds of dimensions
-            'amul': 1, # Alpha multiplier value (wrt to product of mean axes)
+            'a': None,        # Alpha value -> at large alpha,
+                              # hull becomes convex. if None: use product of bounds range
+            'amul': 1,        # Alpha multiplier value (wrt to a used)
+            'opt_amul': True, # If True, use bisection to find largest amul with no vio
+                              # in DSp
+            'maxiter': 50,    # Maximum bisection iterations
+            'tol': 0.001,     # Bisection tolerance (on the MV which is amul)
+            'lb': 0.001,      # Lower bound of initial bisection run
+            'ub': 5,          # Upper bound of initial bisection run
+            'printF': False,  # If true, print iter details
+            'maxvp': 0,       # Maximum allowed percentage of vio in DSp
         }
         self.opt = self.default_opt.copy()
         self.bw_template = {'alpha': 1, 'satcolor': 'gray', 'viocolor': 'black', \
@@ -327,23 +339,17 @@ class DSI():
         # ----- Design space surface/boundary ----- #
         if opt['hidedsp'] == False:
             if self.space_size == None:
-                self.space_size, bF, P, self.shp = self.find_DSp(vnames, opt)
-            bF = self.bF
-            P = self.P
+                self.shp = self.find_DSp(vnames, opt)
+            shp = self.shp
             if dim == 2:
-                for i in range(bF.shape[0]):
-                    if i == 0:
-                        plt.plot(P[bF][i][:, 0], P[bF][i][:, 1],\
-                            linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
-                            color = opt['dspcolor'], label = opt['dsplabel'],\
-                            zorder = opt['dspzorder'])
-                    else:
-                        plt.plot(P[bF][i][:, 0], P[bF][i][:, 1],\
-                            linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
-                            color = opt['dspcolor'], zorder = opt['dspzorder'])
+                plt.plot(*zip(*shp['edges_val']),\
+                    linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
+                    color = opt['dspcolor'], label = opt['dsplabel'],\
+                    zorder = opt['dspzorder'])
 
             if dim == 3:
-                surf = ax.plot_trisurf(*zip(*P), triangles = bF, color = opt['dspcolor'],\
+                surf = ax.plot_trisurf(*zip(*shp['P']), triangles = shp['tri'], \
+                    color = opt['dspcolor'],\
                     alpha = opt['dspalpha'], label = opt['dsplabel'],\
                     zorder = opt['dspzorder'])
                 surf._facecolors2d=surf._facecolor3d
@@ -391,56 +397,236 @@ class DSI():
     
     def find_DSp(self, vnames, opt = {}):
         """
-        Create a hull using MATLAB alphashape
+        Create a hull using alphashape
         Depending on alpha (opt['a']) value it can be either convex/concave
         # ----- Hull Parameters ----- #
-        'a': None, # Alpha value -> at large alpha, hull becomes convex.
-        If set to 'critical', MATLAB finds the smallest one. if None: use mean of bounds 
-        of dimensions
-        'amul': 0.5, # Alpha multiplier value (wrt to product of mean axes)
+        'a': None, # Alpha value -> at large alpha, hull becomes convex
+        If None: use the product of axis bounds range
+        'amul': 2, # Alpha multiplier value (wrt to product of mean axes)
         """
-        import matlab.engine
         import numpy as np
         import pandas as pd
-        eng = matlab.engine.start_matlab() # Start instance of matlab engine
-        ev = eng.eval
-        ew = eng.workspace
         
         sat = self.sat
-        vio = self.sat
+        points = sat[vnames].to_numpy()
+        vio = self.vio
+        vpoints = vio[vnames].to_numpy()
         self.opt.update({'vnames': vnames})
         self.opt.update(opt) 
         opt = self.opt
         a = opt['a']
-        amul = opt['amul']
-        points = sat[vnames].to_numpy()
-        
-        ew['points'] = matlab.double(points.tolist()) 
-        if a == 'critical':
-            shp = ev(f"alphaShape(points)", nargout = 1)
-        elif a == None:
-            a = pd.concat([sat, vio], axis = 0)[vnames].mean().product()*amul
-            shp = ev(f"alphaShape(points, {a})", nargout = 1)
-        else:
-            shp = ev(f"alphaShape(points, {a})", nargout = 1)
-        ew['shp'] = shp
-        
-        if points.shape[1] == 2:
-            space_size = eng.area(shp)
-        elif points.shape[1] == 3:
-            space_size = eng.volume(shp)
-        
-        bF, P = eng.boundaryFacets(shp, nargout = 2)
-        bF = np.array(bF).astype('int') - 1
-        P = np.array(P)
-        
+        if a == None:
+            fdf = pd.concat([sat, vio], axis = 0)[vnames]
+            a = np.product(fdf[vnames].max() - fdf[vnames].min())
+
+        dim = len(vnames)
+        if dim == 2:
+            find_shp = self.alphashape_2D
+            inside = self.inside2D
+        elif dim == 3:
+            find_shp = self.alphashape_3D
+            inside = self.inside3D
+
+        if opt['opt_amul'] == False: # Single run
+            amul = opt['amul']
+            alpha = a*amul
+            self.alpha = alpha
+            shp = find_shp(points, alpha)
+            shp['amul'] = amul
+            sol_flag = f'No optimisation performed'
+
+        elif opt['opt_amul'] == True: # Use bisection
+            opt_log = []
+            tol = opt['tol']
+            lb = opt['lb']
+            print_flag = opt['printF']
+            ub = opt['ub']
+            maxiter = opt['maxiter']
+            maxvnum = opt['maxvp']*self.df.shape[0]
+            # Bisection algorithm
+            for i in range(maxiter):
+                mp = lb + (ub - lb)/2
+                shp = find_shp(points, a*mp)
+                shp['amul'] = mp
+                shp['iter'] = i + 1
+                opt_log.append(shp)
+                self.shp = shp
+                r = inside(vpoints, shp)
+                flag = True in r
+                vnum = vio[r].shape[0]
+                if vnum <= maxvnum: # early break based on maxvnum
+                    if print_flag:
+                        print(i + 1, mp, flag)
+                    sol_flag = f'vnum is LE maxvnum ({opt["maxvp"]*100:.1f}%): {vnum} <= {maxvnum}'
+                    break
+                elif flag == True:
+                    ub = mp
+                elif flag == False:
+                    lb = mp
+                    if ub - lb <= tol:
+                        if print_flag:
+                            print(i + 1, mp, flag)
+                        sol_flag = f'[{i + 1}] Optimal amul found: {mp:6f}   Gap: {ub - lb:6f}'
+                        break
+                if print_flag:
+                    print(i + 1, mp, flag)
+            if (i + 1) == maxiter:
+                sol_flag = f'Max iterations reached: {maxiter} iterations  amul: {mp:6f}'
+        if print_flag:
+            print(sol_flag)
+
+        space_size = shp['size']
         self.shp = shp
         self.space_size = space_size
-        self.bF = bF
-        self.P = P
+        self.bF = shp['tri']
+        self.P = shp['P']
         self.report['space_size'] = space_size
-        return space_size, bF, P, shp
+        self.report['sol_flag'] = sol_flag
+        self.report['opt_log'] = opt_log
+        return shp
     
+    def alphashape_2D(self, P, alpha):
+        """
+        Calculate the alphashape boundary from a point cloud P (2D np array)
+        Adapted from Kostas Markakis (https://stackoverflow.com/users/10105748/kostas-markakis)
+        https://stackoverflow.com/a/62951837
+        """
+        from shapely.geometry import MultiLineString
+        from shapely.ops import unary_union, polygonize
+        from scipy.spatial import Delaunay
+        from collections import Counter
+        import numpy as np
+        import itertools
+
+        v = Delaunay(P).vertices
+        # Calculate the sides of the triangles
+        a = (P[v][:, 0][:, 0] - P[v][:, 1][:, 0])**2 + \
+            (P[v][:, 0][:, 1] - P[v][:, 1][:, 1])**2
+        b = (P[v][:, 1][:, 0] - P[v][:, 2][:, 0])**2 + \
+            (P[v][:, 1][:, 1] - P[v][:, 2][:, 1])**2
+        c = (P[v][:, 2][:, 0] - P[v][:, 0][:, 0])**2 + \
+            (P[v][:, 2][:, 1] - P[v][:, 0][:, 1])**2
+        a, b, c = np.sqrt([a, b, c])
+        s = (a + b + c)*0.5
+        area = np.sqrt(s*(s - a)*(s - b)*(s - c)) # area from Heron's formula
+        alpha_filter = a*b*c/(4.0*area) < alpha
+
+        # Filter vertices of alpha shape based on alpha radius
+        tri = v[alpha_filter]
+        # edges = np.vstack([edges[:, 0:2], edges[:, 1:3], np.delete(edges, 2, 1)])
+        # edges = list(zip(edges[:, 0], edges[:, 1]))
+        edges = [tuple(sorted(cb)) for e in tri for cb in itertools.combinations(e, 2)]
+
+        count = Counter(edges)
+        # Keep only edges that appear one time (concave hull edges)
+        edges = [e for e, c in count.items() if c == 1]
+        edges_val = [(P[e[0]], P[e[1]]) for e in edges]
+
+        # Return points in order for plotting
+        ml = MultiLineString(edges_val)
+        poly = polygonize(ml)
+        hull = unary_union(list(poly))
+        hull_vertices = hull.exterior.coords.xy
+        h = np.zeros((len(hull_vertices[0]), 2))
+        h[:, 0] = hull_vertices[0]
+        h[:, 1] = hull_vertices[1]
+
+        shp = {}
+        shp['P'] = P
+        shp['verts'] = np.unique(edges)
+        shp['tri'] = tri
+        shp['tetras'] = None
+        shp['edges'] = edges
+        shp['edges_val'] = h
+        shp['alpha'] = alpha
+
+        # Area of alpha shape
+        v = tri
+        a = (P[v][:, 0][:, 0] - P[v][:, 1][:, 0])**2 + \
+            (P[v][:, 0][:, 1] - P[v][:, 1][:, 1])**2
+        b = (P[v][:, 1][:, 0] - P[v][:, 2][:, 0])**2 + \
+            (P[v][:, 1][:, 1] - P[v][:, 2][:, 1])**2
+        c = (P[v][:, 2][:, 0] - P[v][:, 0][:, 0])**2 + \
+            (P[v][:, 2][:, 1] - P[v][:, 0][:, 1])**2
+        a, b, c = np.sqrt([a, b, c])
+
+        s = (a + b + c)*0.5
+        area = np.sqrt(s*(s - a)*(s - b)*(s - c)) # area from Heron's formula
+        shp['size'] = np.sum(area)
+        return shp
+
+    def alphashape_3D(self, P, alpha):
+        """
+        Calculate the alphashape boundary from a point cloud P (3D np array)
+        Adapted from Geun (https://stackoverflow.com/users/9091202/geun)
+        https://stackoverflow.com/a/58113037
+
+        Compute the alpha shape (concave hull) of a set of 3D points.
+        Parameters:
+            pos - np.array of shape (n,3) points.
+            alpha - alpha value.
+        return
+            outer surface vertex indices, edge indices, and triangle indices
+        """
+        from scipy.spatial import Delaunay
+        import numpy as np
+        from collections import defaultdict
+
+        tetra = Delaunay(P)
+        # Find radius of the circumsphere.
+        # By definition, radius of the sphere fitting inside the tetrahedral needs 
+        # to be smaller than alpha value
+        # http://mathworld.wolfram.com/Circumsphere.html
+
+        tetrapos = np.take(P, tetra.vertices,axis=0)
+        normsq = np.sum(tetrapos**2, axis=2)[:, :, None]
+        ones = np.ones((tetrapos.shape[0], tetrapos.shape[1], 1))
+
+        a  =  np.linalg.det(np.concatenate((tetrapos, ones), axis=2))
+        a[a == 0] = 1e-30
+        Dx =  np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[1,2]], ones), axis = 2))
+        Dy = -np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[0,2]], ones), axis = 2))
+        Dz =  np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[0,1]], ones), axis = 2))
+        c  =  np.linalg.det(np.concatenate((normsq, tetrapos), axis = 2))
+        r  =  np.sqrt(Dx**2 + Dy**2 + Dz**2 - 4*a*c)/(2*np.abs(a))
+
+        # Find tetrahedrals
+        tetras = tetra.vertices[r<alpha, :]
+
+        # triangles
+        triComb = np.array([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)])
+        triangles = tetras[:, triComb].reshape(-1, 3)
+        triangles = np.sort(triangles, axis = 1)
+
+        # Remove triangles that occurs twice, because they are within shapes
+        trianglesDict = defaultdict(int)
+        for tri in triangles:
+            trianglesDict[tuple(tri)] += 1
+        triangles = np.array([tri for tri in trianglesDict if trianglesDict[tri] == 1])
+
+        #edges
+        edgeComb = np.array([(0, 1), (0, 2), (1, 2)])
+        edges = triangles[:, edgeComb].reshape(-1, 2)
+        edges = np.sort(edges, axis = 1)
+        edges = np.unique(edges, axis = 0)
+
+        vertices = np.unique(edges)
+
+        shp = {}
+        shp['P'] = P
+        shp['verts'] = vertices
+        shp['tri'] = triangles
+        shp['edges'] = edges
+        shp['tetras'] = tetras
+        shp['edges_val'] = None
+        shp['alpha'] = alpha
+
+        volume = 0
+        for v in P[tetras]:
+            volume += np.abs(np.dot(v[0] - v[3], np.cross(v[1] - v[3], v[2] - v[3])))/6
+        shp['size'] = volume
+        return shp
+
     def collect_frames(self, anielev = 10, sfolder = 'animation', sname = 'plot',\
          leading_no = 0, sdpi = 100):
         """
@@ -469,135 +655,181 @@ class DSI():
                 ax.azim += 0.5
                 plt.savefig(f'{sfolder}{sname}_frame_{leading_no + ii:04}.png',\
                     dpi = sdpi)
+    def inside2D(self, x, shp = None):
+        """
+        Returns False if x is not in self.shp, True otherwise (2D)
+        x: list of [x, y], or [[x1, y1], [x2, y2], ...]
+        """
+        import numpy as np
+        from shapely.geometry import Point, Polygon
+        if shp == None:
+            shp = self.shp
+        poly = Polygon(shp['edges_val'])
+        dim = len(np.array(x).shape)
+
+        if dim == 1:
+            x = [x]
+        vP = [Point(i[0], i[1]) for i in x]
+        out = [poly.contains(p) for p in vP]
+
+        if dim == 1:
+            out = out[0]
+        return out
     
+    def inside3D(self, x, shp = None):
+        """
+        Returns False if x is not in self.shp, True otherwise (3D)
+        x: list of [x, y, z], or [[x1, y1, z1], [x2, y2, z2], ...]
+        https://stackoverflow.com/a/57901916
+        https://stackoverflow.com/a/41851137/12056867
+        """
+        import numpy as np
+        if shp == None:
+            shp = self.shp
+        dim = len(np.array(x).shape)
+
+        # Find which tetrahedron the point lies in
+        node_coordinates = shp['P']
+        node_ids = shp['tetras']
+        p = np.array(x)
+
+        ori = node_coordinates[node_ids[:, 0],:]
+        v1 = node_coordinates[node_ids[:, 1],:] - ori
+        v2 = node_coordinates[node_ids[:, 2],:] - ori
+        v3 = node_coordinates[node_ids[:, 3],:] - ori
+        n_tet = len(node_ids)
+        v1r = v1.T.reshape((3, 1, n_tet))
+        v2r = v2.T.reshape((3, 1, n_tet))
+        v3r = v3.T.reshape((3, 1, n_tet))
+        mat = np.concatenate((v1r, v2r, v3r), axis=1)
+        inv_mat = np.linalg.inv(mat.T).T
+        if p.size == 3:
+            p = p.reshape((1,3))
+        n_p = p.shape[0]
+        orir = np.repeat(ori[:,:,np.newaxis], n_p, axis=2)
+        newp = np.einsum('imk,kmj->kij',inv_mat,p.T-orir)
+        val = np.all(newp>=0, axis=1) & np.all(newp <=1, axis=1) & (np.sum(newp, axis=1)<=1)
+        id_tet, id_p = np.nonzero(val)
+        res = -np.ones(n_p, dtype=id_tet.dtype) # Sentinel value
+        res[id_p]=id_tet
+        
+        if dim == 1:
+            x = [x]
+        # return res
+        out = []
+        for i, r in enumerate(res):
+            V = shp['P'][shp['tetras'][r]]
+            p = x[i]
+            # Find the transform matrix from orthogonal to tetrahedron system
+            v1 = V[1]-V[0] ; v2 = V[2]-V[0] ; v3 = V[3]-V[0]
+            mat = np.array((v1,v2,v3)).T
+            # mat is 3x3 here
+            M1 = np.linalg.inv(mat)
+            # apply the transform to P (v1 is the origin)
+            newp = M1.dot(p - V[0])
+            # perform test
+            out.append(np.all(newp>=0) and np.all(newp <=1) and np.sum(newp)<=1)
+        if dim == 1:
+            out = out[0]
+        return out
+
     def check_point(self, x):
         """
         Checks whether point x lies within the hull or not.
         """
         import numpy as np
-        import matlab.engine
-        eng = matlab.engine.start_matlab() # Start instance of matlab engine
-        
-        shp = self.shp
-        vnames = self.vnames
+        vn = self.vnames
         opt = self.opt
         sat = self.sat
-        
-        dim = len(vnames)
-        step_change = opt['step_change']
         df = self.df
-                        
-        # Finding space boundary
-        inputs_max = df[vnames].max().to_numpy()
-        inputs_min = df[vnames].min().to_numpy()
+
+        # Bisection params
+        pclb = opt['AORlb']
+        pcub = opt['AORub']
+        tol = opt['AORtol']
+        maxiter = opt['AORmaxiter']
+        print_flag = opt['AORprintF']
+        AORopt_log = []
+
+        dim = len(vn)
+        if dim == 2:
+            inside = self.inside2D
+        elif dim == 3:
+            inside = self.inside3D
+
+        inputs_max = df[vn].max().to_numpy()
+        inputs_min = df[vn].min().to_numpy()
         inputs_range = inputs_max - inputs_min
-        pc = step_change/100
-        
+
+
         not_in_region_flag = False
-        if eng.inShape(shp, matlab.double(list(x))) == False:
+        if inside(x) == False:
             print('x is not inside DSp.')
             not_in_region_flag = True
             fs_R = {'x': x, 'FR': 'N/A', 'rmax': 'N/A', 'rmin': 'N/A',
             'space_size': 'N/A', 'plusmin': 'N/A', 'nosam': 'N/A', 
             'hmv': 'N/A', 'hmv_sam_flag': 'N/A', 'not_in_region_flag': not_in_region_flag}
-        else:
-            # ----- 2D space ----- #
+        else: # Bisection
+            for i in range(maxiter):
+                pc = pclb + (pcub - pclb)/2
+                gap = pcub - pclb
+                if dim == 2:
+                    verts = np.array(
+                            [[x[0] - pc*inputs_range[0], x[1] - pc*inputs_range[1]],
+                            [x[0] - pc*inputs_range[0], x[1] + pc*inputs_range[1]],
+                            [x[0] + pc*inputs_range[0], x[1] + pc*inputs_range[1]],
+                            [x[0] + pc*inputs_range[0], x[1] - pc*inputs_range[1]]]
+                                )
+                elif dim == 3:
+                    verts = np.array(
+                            [[x[0] - pc*inputs_range[0],\
+                            x[1] - pc*inputs_range[1], x[2] - pc*inputs_range[2]],
+                            [x[0] + pc*inputs_range[0],\
+                            x[1] - pc*inputs_range[1], x[2] - pc*inputs_range[2]],
+                            [x[0] - pc*inputs_range[0],\
+                            x[1] + pc*inputs_range[1], x[2] - pc*inputs_range[2]],
+                            [x[0] - pc*inputs_range[0],\
+                            x[1] - pc*inputs_range[1], x[2] + pc*inputs_range[2]],
+                            [x[0] + pc*inputs_range[0],\
+                            x[1] + pc*inputs_range[1], x[2] - pc*inputs_range[2]],
+                            [x[0] + pc*inputs_range[0],\
+                            x[1] + pc*inputs_range[1], x[2] + pc*inputs_range[2]],
+                            [x[0] + pc*inputs_range[0],\
+                            x[1] - pc*inputs_range[1], x[2] + pc*inputs_range[2]],
+                            [x[0] - pc*inputs_range[0],\
+                            x[1] + pc*inputs_range[1], x[2] + pc*inputs_range[2]]]
+                                )
+                flag = False in [inside(verts[i]) for i in range(verts.shape[0])]
+                AORopt_log.append({'iter': i+1, 'pc': pc, 'gap': gap, 'flag': flag, 'verts': verts})
+                if flag == False:
+                    pclb = pc
+                    if gap <= tol:
+                        if print_flag:
+                            print(i+1, flag, gap)
+                        break
+                elif flag == True:
+                    pcub = pc
+                if print_flag:
+                    print(i+1, flag, gap)
+
             if dim == 2:
-                flag = False
-                while flag == False: # Going outwards
-                    fsi   = np.array(
-                        [[x[0] - pc*inputs_range[0], x[1] - pc*inputs_range[1]],
-                        [x[0] - pc*inputs_range[0], x[1] + pc*inputs_range[1]],
-                        [x[0] + pc*inputs_range[0], x[1] + pc*inputs_range[1]],
-                        [x[0] + pc*inputs_range[0], x[1] - pc*inputs_range[1]]]
-                        )
-                    flag = False in [eng.inShape(shp, matlab.double(list(fsi[i])))\
-                        for i in range(fsi.shape[0])]
-                    pc += step_change/100
-
-                pc = (step_change/100)/100
-                flag = True
-                while flag == True: # Going inwards
-                    fs = [
-                        [fsi[0, 0] + pc*inputs_range[0], fsi[0, 1] + pc*inputs_range[1]],
-                        [fsi[1, 0] + pc*inputs_range[0], fsi[1, 1] - pc*inputs_range[1]],
-                        [fsi[2, 0] - pc*inputs_range[0], fsi[2, 1] - pc*inputs_range[1]],
-                        [fsi[3, 0] - pc*inputs_range[0], fsi[3, 1] + pc*inputs_range[1]]
-                        ]
-                    fs = np.array(fs)
-                    flag = False in [eng.inShape(shp, matlab.double(list(fs[i])))\
-                        for i in range(fs.shape[0])]
-                    pc += (step_change/100)/100
-                    
-                fs = fs.tolist()
-                fs.append(fs[0])
-                fs_arr = np.array(fs)
-                FR = np.array(fs)
-                
-            # ----- 3D space ----- #
-            if dim == 3: 
-                flag = False
-                while flag == False: # Going outwards
-                    fsi = np.array(
-                        [[x[0] - pc*inputs_range[0],\
-                        x[1] - pc*inputs_range[1], x[2] - pc*inputs_range[2]],
-                        [x[0] + pc*inputs_range[0],\
-                        x[1] - pc*inputs_range[1], x[2] - pc*inputs_range[2]],
-                        [x[0] - pc*inputs_range[0],\
-                        x[1] + pc*inputs_range[1], x[2] - pc*inputs_range[2]],
-                        [x[0] - pc*inputs_range[0],\
-                        x[1] - pc*inputs_range[1], x[2] + pc*inputs_range[2]],
-                        [x[0] + pc*inputs_range[0],\
-                        x[1] + pc*inputs_range[1], x[2] - pc*inputs_range[2]],
-                        [x[0] + pc*inputs_range[0],\
-                        x[1] + pc*inputs_range[1], x[2] + pc*inputs_range[2]],
-                        [x[0] + pc*inputs_range[0],\
-                        x[1] - pc*inputs_range[1], x[2] + pc*inputs_range[2]],
-                        [x[0] - pc*inputs_range[0],\
-                        x[1] + pc*inputs_range[1], x[2] + pc*inputs_range[2]]]
-                                    )
-                    flag = False in [eng.inShape(shp, matlab.double(list(fsi[i])))\
-                        for i in range(fsi.shape[0])]
-                    pc += step_change/100
-
-                pc = (step_change/100)/100
-                flag = True
-                while flag == True: # Going inwards
-                    fs = [[fsi[0, 0] + pc*inputs_range[0],\
-                         fsi[0, 1] + pc*inputs_range[1], fsi[0, 2] + pc*inputs_range[2]],
-                        [fsi[1, 0] - pc*inputs_range[0],\
-                        fsi[1, 1] + pc*inputs_range[1], fsi[1, 2] + pc*inputs_range[2]],
-                        [fsi[2, 0] + pc*inputs_range[0],\
-                        fsi[2, 1] - pc*inputs_range[1], fsi[2, 2] + pc*inputs_range[2]],
-                        [fsi[3, 0] + pc*inputs_range[0],\
-                        fsi[3, 1] + pc*inputs_range[1], fsi[3, 2] - pc*inputs_range[2]],
-                        [fsi[4, 0] - pc*inputs_range[0],\
-                        fsi[4, 1] - pc*inputs_range[1], fsi[4, 2] + pc*inputs_range[2]],
-                        [fsi[5, 0] - pc*inputs_range[0],\
-                        fsi[5, 1] - pc*inputs_range[1], fsi[5, 2] - pc*inputs_range[2]],
-                        [fsi[6, 0] - pc*inputs_range[0],\
-                        fsi[6, 1] + pc*inputs_range[1], fsi[6, 2] - pc*inputs_range[2]],
-                        [fsi[7, 0] + pc*inputs_range[0],\
-                        fsi[7, 1] - pc*inputs_range[1], fsi[7, 2] - pc*inputs_range[2]]]
-                    fs = np.array(fs)
-                    fs_arr = fs
-                    flag = False in [eng.inShape(shp, matlab.double(list(fs[i])))\
-                        for i in range(fs.shape[0])]
-                    pc += (step_change/100)/100
-                    
-                FR = [[fs[0], fs[1], fs[4], fs[2], fs[0]],
-                        [fs[0], fs[3], fs[6], fs[1], fs[0]],
-                        [fs[3], fs[7], fs[5], fs[6], fs[3]], 
-                        [fs[6], fs[1], fs[4], fs[5], fs[6]],
-                        [fs[7], fs[3], fs[0], fs[2], fs[7]],
-                        [fs[2], fs[4], fs[5], fs[7], fs[2]]]
+                verts = verts.tolist()
+                verts.append(verts[0])
+                FR = np.array(verts)
+            elif dim == 3:
+                FR = [[verts[0], verts[1], verts[4], verts[2], verts[0]],
+                    [verts[0], verts[3], verts[6], verts[1], verts[0]],
+                    [verts[3], verts[7], verts[5], verts[6], verts[3]], 
+                    [verts[6], verts[1], verts[4], verts[5], verts[6]],
+                    [verts[7], verts[3], verts[0], verts[2], verts[7]],
+                    [verts[2], verts[4], verts[5], verts[7], verts[2]]]
 
             # ----- KPIs ----- #
-            rmax = fs_arr.max(axis = 0)
-            rmin = fs_arr.min(axis = 0)
+            rmax = np.array(verts).max(axis = 0)
+            rmin = np.array(verts).min(axis = 0)
             fs_size = (rmax - rmin).prod()
             plusmin = (rmax - rmin)/2
-            
+
             # ----- Heat map variable ----- #
             hmv_fs_R = {'name': opt['hmv'], 'mean': '-', 'max': '-', 'max_sample': '-', 
             'min': '-',  'min_sample': '-', 'fs_all_samples': '-'}
@@ -606,9 +838,9 @@ class DSI():
             if opt['hmv'] != 'None':
                 fs_df = sat.copy()
                 for i in range(dim):
-                    fs_df = fs_df[fs_df[vnames[i]] <= rmax[i]]
+                    fs_df = fs_df[fs_df[vn[i]] <= rmax[i]]
                 for i in range(dim):
-                    fs_df = fs_df[fs_df[vnames[i]] >= rmin[i]]
+                    fs_df = fs_df[fs_df[vn[i]] >= rmin[i]]
                 if fs_df.shape[0] == 0:
                     print('No samples inside AOR available.')
                     no_samples_flag = True
@@ -621,8 +853,8 @@ class DSI():
                     hmv_fs_R['fs_all_samples'] = fs_df
             fs_R = {'x': x, 'FR': FR, 'rmax': rmax, 'rmin': rmin, 'space_size': fs_size, 
                 'plusmin': plusmin, 'nosam': fs_df.shape[0], 'hmv': hmv_fs_R, 
-                'hmv_sam_flag': no_samples_flag, 'not_in_region_flag': not_in_region_flag}
-                
+                'hmv_sam_flag': no_samples_flag, 'not_in_region_flag': not_in_region_flag, 'AORopt_log': AORopt_log}
+
         self.all_x.update({str(x): fs_R})
         return fs_R
         
